''
'' AVI Player (FreeBasic)
'' by ssjx (http://ssjx.co.uk)
''
'' This program can play AVI files created using VirtualDub and possibly other applications (untested).
''
'' The AVI needs to:
''		* Have a resolutions of less than 320x240
''		* Have a width that is a multiple of 4
''		* Be either Uncompressed or compressed using the MSVC codec.
''		* Have no audio
''
'' If the AVI is uncompressed and has a resolution <= 160x120, it will be doubled.
''
'' Compile with:
''		fbc *.bas
''
'' For additonal checks/safety, compile with: 
''		fbc -exx iff.bas
''

#include "avi.bi"

#define SC_ESCAPE &h01

const TITLE="AVI Player"
const VERSION="v0.1 (##/08/2020)"
const SITEURL="http://ssjx.co.uk"

dim	aheader as myavih
dim	vheader as myvshdr
dim	vformat as myvstrf	
dim	chunk as mychunk
dim	header as myheader

Dim shared bmp As ubyte Ptr

dim shared as boolean scale=false

''
sub decode_dib(w as integer,h as integer)
	dim as integer px=0
	
	for y as integer=h-1 to 0 step -1
		for x as integer=0 to w-1
			
			if scale=false then
				pset (x,y),rgb(bmp[px+2],bmp[px+1],bmp[px])
			else
				line (x*2,y*2)-((x*2)+1,(y*2)+1),rgb(bmp[px+2],bmp[px+1],bmp[px]),bf
			end if
			px+=3
		next x
	next y
end sub

'' 16bit to 24bit RGB
function col(incol as ushort) as uinteger
	dim as uinteger r=((incol and 31)) shl 3
	dim as uinteger g=((incol and 992) shr 5) shl 3
	dim as uinteger b=((incol and &h7c00) shr 10) shl 3
	
	return rgb(b,g,r)
end function

sub decode_msvc(w as integer,h as integer,datasize as integer)
	dim as ubyte code0,code1,flag
	dim as ushort ca,cb,cc,cd

	dim as boolean quit=false
	dim as integer x=0,y=h-1,skip,ps=0,ofy

	while quit=false
	    code0=bmp[ps]
	    code1=bmp[ps+1]
	
		ps+=2
		if ps>datasize or y<=0 then return
	
		if ((code1>=&h84) and (code1<=&h87)) then
	      	skip=((code1-&h84) shl 8)+code0
	      	
			while(skip>0)
				skip-=1
				x+=4
				if(x>=w) then 
					y-=4
					x=0
				end if
			wend
	
	    else
	    	if (code1<&h80) then
				ca=bmp[ps+0]+(bmp[ps+1] shl 8)
				cb=bmp[ps+2]+(bmp[ps+3] shl 8)
				
				if (ca and &h8000) then
					'' 4x4  with 4 colours
					cc=bmp[ps+4]+(bmp[ps+5] shl 8)
					cd=bmp[ps+6]+(bmp[ps+7] shl 8)	
					
					flag=code0
					ofy=0
					for r as integer=0 to 1
						if (flag and &h01)=&h01 then pset(x+0,y-ofy),col(ca) else pset(x+0,y-ofy),col(cb) 
						if (flag and &h02)=&h02 then pset(x+1,y-ofy),col(ca) else pset(x+1,y-ofy),col(cb) 
						if (flag and &h04)=&h04 then pset(x+2,y-ofy),col(cc) else pset(x+2,y-ofy),col(cd) 
						if (flag and &h08)=&h08 then pset(x+3,y-ofy),col(cc) else pset(x+3,y-ofy),col(cd)
						ofy+=1
						if (flag and &h10)=&h10 then pset(x+0,y-ofy),col(ca) else pset(x+0,y-ofy),col(cb)
						if (flag and &h20)=&h20 then pset(x+1,y-ofy),col(ca) else pset(x+1,y-ofy),col(cb)
						if (flag and &h40)=&h40 then pset(x+2,y-ofy),col(cc) else pset(x+2,y-ofy),col(cd)
						if (flag and &h80)=&h80 then pset(x+3,y-ofy),col(cc) else pset(x+3,y-ofy),col(cd)
						ofy+=1
						
						if r=0 then
							ca=bmp[ps+8]+(bmp[ps+9] shl 8)
							cb=bmp[ps+10]+(bmp[ps+11] shl 8)
							cc=bmp[ps+12]+(bmp[ps+13] shl 8)
							cd=bmp[ps+14]+(bmp[ps+15] shl 8)
							flag=code1
						end if
					next r		
					ps+=12
				else
					'' 4x4  with 2 colours
					flag=code0
					ofy=0
					for r as integer=0 to 1
						if (flag and &h01)=&h01 then pset(x+0,y-ofy),col(ca) else pset(x+0,y-ofy),col(cb)
						if (flag and &h02)=&h02 then pset(x+1,y-ofy),col(ca) else pset(x+1,y-ofy),col(cb)
						if (flag and &h04)=&h04 then pset(x+2,y-ofy),col(ca) else pset(x+2,y-ofy),col(cb)
						if (flag and &h08)=&h08 then pset(x+3,y-ofy),col(ca) else pset(x+3,y-ofy),col(cb)
						ofy+=1
						if (flag and &h10)=&h10 then pset(x+0,y-ofy),col(ca) else pset(x+0,y-ofy),col(cb)
						if (flag and &h20)=&h20 then pset(x+1,y-ofy),col(ca) else pset(x+1,y-ofy),col(cb)
						if (flag and &h40)=&h40 then pset(x+2,y-ofy),col(ca) else pset(x+2,y-ofy),col(cb)
						if (flag and &h80)=&h80 then pset(x+3,y-ofy),col(ca) else pset(x+3,y-ofy),col(cb)
						ofy+=1
						flag=code1
					next r
				end if
				
				ps+=4
			else
				'' 4x4 with 1 colour
				ca=code0+(code1 shl 8)
				for j as integer=0 to 3
					for i as integer=0 to 3 
						pset(x+i,y-j),col(ca)
					next i
				next j		
			end if
			
			x+=4
			if(x>=w) then
				y-=4
				x=0
			end if
	    end if
	wend
end sub



''
''
dim as string myfile
myfile = Command(1)

if myfile="" then
	print "Usage : avi myfile.avi"
	end 
else
	print "Loading file "+myfile
end if

if  Open(myfile For binary access read As #1)=0 then
	
	dim as string stype,ctype,compression
	dim as integer fpos=13,nsize
	dim as boolean ismovie=false
	dim as integer fps
	
	for i as integer=0 to 3
		get #1,fpos,chunk	 
		
		stype=chr(chunk.subtype(0))+chr(chunk.subtype(1))+chr(chunk.subtype(2))+chr(chunk.subtype(3))	
		nsize=chunk.size
		'print stype
		
		if stype="hdrl" then
			'' AVI file header
			get #1,,header	
			get #1,,aheader	
			print "Frames: "+str(aheader.totf)
		
			get #1,,chunk	 
			ctype=chr(chunk.subtype(0))+chr(chunk.subtype(1))+chr(chunk.subtype(2))+chr(chunk.subtype(3))	
			'print "Desc: "+ctype
	
			for c as integer=0 to 1
				get #1,,header	 
				ctype=chr(header.db(0))+chr(header.db(1))+chr(header.db(2))+chr(header.db(3))	
			
				select case ctype
				case "strh"
					get #1,,vheader
					compression=chr(vheader.shand(0))+chr(vheader.shand(1))+chr(vheader.shand(2))+chr(vheader.shand(3))	
					compression=ucase(compression)
					print "Format: ["+compression+"]"
					fps=vheader.fps/ vheader.scale
					print "Rate: "+str(vheader.fps)+" / "+str(vheader.scale)+" ("+str(fps)+")"
					
				case "strf"
					get #1,,vformat
					print "Size: "+str(vformat.xsize)+" x "+str(vformat.ysize)+" x "+str(vformat.depth)
						
				end select
			
			next c
	
		end if		
		
		if stype="movi" then 
				ismovie=true
				exit for
		end if
		
		fpos+=(nsize+8)
	next i	
			
	''
	if compression<>"DIB " and compression<>"MSVC" then
		print "Unsupported compression ["+compression+"], video must be either uncompressed or MSVC.";
		ismovie=false
	end if
	
	if (vformat.xsize>320) or  (vformat.ysize>240) then
		print "Resolution must be 320x240 or lower."
		ismovie=false
	end if
	
	if (vformat.xsize mod 4)>0 then
		print "AVI width must be divisible by 4"
		ismovie=false
	end if
	
	if ismovie=true then
		
		bmp = CAllocate(320*240*4)
		''
		'' Trim off the full path for the window display
		''
		dim as integer idx
		idx=InStrRev(myfile,"\")

		dim as string nicename
		if idx=0 then
			nicename=myfile
		else
			nicename=right(myfile,len(myfile)-idx)
		end if
		
		if vformat.xsize<=160 and vformat.ysize<=120 then scale=true
		
		''
		screen 14,32,1 '' 18
		windowtitle nicename+" - "+TITLE+" "+VERSION+" - "+SITEURL
		
		for i as integer=0 to aheader.totf-1 
			'' Read a frame from the AVI
			get #1,,header
			'stype=chr(header.db(0))+chr(header.db(1))+chr(header.db(2))+chr(header.db(3))
			'print "Desc : "+stype+" Size:"+str(header.size)
			
			get #1,,*bmp,header.size
			
			'' Decode and show frame
			screenlock
				if compression="DIB " then decode_dib(vformat.xsize,vformat.ysize)
				if compression="MSVC" then decode_msvc(vformat.xsize,vformat.ysize,header.size)
			screenunlock
			
			if multikey(SC_ESCAPE) then exit for

			sleep (1000/(fps+2)),1	
		next i
		
		Deallocate(bmp)
	end if
	
	Close #1
else
	print "Can't find "+myfile
end  if
	
	
	
	